package fr.doap.trash;

import fr.doap.jdb.*;
import fr.doap.slog.*;

import java.sql.*;
import java.util.*;
import java.util.function.*;

public class DBWrapper {
	
	DBWrapper(String aConn, String aUser, String aPass) {
		try {
			if (mConnection != null) mConnection.close();
			mConnection = DriverManager.getConnection(aConn+";DB_CLOSE_ON_EXIT=FALSE", aUser, aPass);
			
			mSQLQueue = new ProcessingQueue<>(aSQL -> aSQL.execute());
			
			Runtime.getRuntime().addShutdownHook(new Thread(() -> {
				L.t("Waiting for queue to finish...");
				mSQLQueue.stop().lock();
				L.t("Queue empty... finishing!");

				try {
					if ((mConnection != null)&&(!mConnection.isClosed())) {
						mConnection.close();
						L.i("Connection closed");
					}
				} catch (Exception aE) { aE.printStackTrace(); }
			}));
			
			ResultSet lRST = mConnection.getMetaData().getTables(null, null, "OBJECTS", null);
			if (!lRST.next()) mSQLQueue.add(new SQLCommand("CREATE TABLE OBJECTS ("+sJID+" INT PRIMARY KEY NOT NULL, class VARCHAR)"));
		
			mSQLQueue.add(new SQLExec(aConnection -> {
				try {
					ResultSet lRSPK = aConnection.getMetaData().getColumns(null, null, "OBJECTS", null);
					L.i("--- Columns:");
					while (lRSPK.next()) {
						L.i("Table: {} / Column: {} / Type: {}", lRSPK.getString(3), lRSPK.getString(4), JDBCType.valueOf(lRSPK.getInt(5)).getName());
					}
					L.i("--- No more columns");
				} catch (Exception aE) { aE.printStackTrace(); }
			}));
//			System.out.println("Registered Primary keys: "+ sPrimaryKeys);
			
		} catch (Exception aE) {
			aE.printStackTrace();
		}
	}
	
	protected Connection							mConnection;
	protected WeakObjectCache						mWeakObjectCache = new WeakObjectCache();
	protected ProcessingQueue<SQLAbstractCommand>	mSQLQueue;
	
	interface SQLAbstractCommand {
		void execute();
	}
	
	class SQLExec implements SQLAbstractCommand {
		SQLExec(Consumer<Connection> aCallback) { mCallback=aCallback; }
		
		@Override
		public void execute() {
			mCallback.accept(mConnection);
		}
		
		Consumer<Connection>	mCallback;
	}
	
	abstract class SQLSimpleCommand implements SQLAbstractCommand {
		abstract protected String getSQL();
		
		public void execute() {
			try {
				String lSQL = getSQL();
				L.t("SQL: {}", lSQL);
				
				mConnection.createStatement().execute(lSQL);
			} catch (Exception aE) {
				aE.printStackTrace();
			}
		}
	}
	
	class SQLCommand extends SQLSimpleCommand {
		SQLCommand(String aS) { mCommand = aS; }
		
		@Override
		protected String getSQL() {	return mCommand; }
		
		String mCommand;
	}
	
	class SQLQuery extends SQLCommand {
		SQLQuery(String aQuery, Consumer<ResultSet> aRS) {
			super(aQuery);
			mResultSetConsumer = aRS;
		}
		
		@Override
		public void execute() {
			try {
				String lSQL = getSQL();
				L.t("SQL: {}", lSQL);
				
				mResultSetConsumer.accept(mConnection.createStatement().executeQuery(lSQL));
			} catch (Exception aE) {
				aE.printStackTrace();
			}
		}
		
		Consumer<ResultSet>	mResultSetConsumer;
	}
	
	abstract class SQLParameterizedAbstractCommand implements SQLAbstractCommand {
		abstract protected String getSQL();
		
		protected SQLParameterizedAbstractCommand(Map<String, Object> aKV) {
			mKeyValues = aKV;
		}
		
		public void execute() {
			try {
				String lSQL = getSQL();
				L.t("SQL: {}", lSQL);

				PreparedStatement lPS = mConnection.prepareStatement(lSQL, new String[]{ sJID });
				List<Object> lCol = mKeyValues.values().stream().toList();
				for (int i = 0; i < lCol.size(); i++) lPS.setObject(i + 1, lCol.get(i));
				lPS.execute();
				ResultSet lRS = lPS.getGeneratedKeys();
				if (lRS.next()) mKeyValues.put(sJID, lRS.getInt(1));
			} catch (Exception aE) {
				aE.printStackTrace();
			}
		}
		
		Map<String, Object>	mKeyValues = new HashMap<>();
	}

	class SQLInsert extends SQLParameterizedAbstractCommand {
		protected SQLInsert(Map<String, Object> aKV) { super(aKV); }
		
		@Override
		protected String getSQL() {
			return "INSERT INTO objects ("+String.join(", ", mKeyValues.keySet())+") VALUES ("+String.join(", ", Collections.nCopies(mKeyValues.size(), "?"))+")";
		}
	}
	
	class SQLUpdate extends SQLParameterizedAbstractCommand {
		protected SQLUpdate(Map<String, Object> aKV) { super(aKV); }
		
		@Override
		protected String getSQL() {
			int lID = (int) mKeyValues.get(sJID);
			mKeyValues.remove(sJID);
			return "UPDATE objects SET "+String.join(", ", mKeyValues.keySet().stream().map(aK -> aK+"=?").toList())+" WHERE "+sJID+"="+lID;
		}
	}

	static String sJID = "_JID_";
}
